class SqlBuilder:
    def __init__(self, table: str):
        self.table = table
        self.basicSelect = f'SELECT * FROM [{table}]'
        self.primaryKeyQuery = f"""
        SELECT COLUMN_NAME AS Column_name
        FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE
        WHERE OBJECTPROPERTY(OBJECT_ID(CONSTRAINT_NAME), 'IsPrimaryKey') = 1
        AND TABLE_NAME = '{table}'
    """

    def insert(self, data):
        if isinstance(data, dict):
            keys = list(data.keys())
        elif isinstance(data, list):
            keys = list(data[0].keys())

        values = []
        for key in keys:
            values.append(f"%({key})s")

        sql = f"INSERT INTO [{self.table}] ({','.join(keys)}) VALUES({','.join(values)})"
        return sql

    def update(self, data: dict):
        pairs = []
        for key in data:
            pairs.append(f"{key} = %({key})s")

        return f"UPDATE [{self.table}] SET {','.join(pairs)}"

    def delete(self):
        return f"DELETE FROM [{self.table}]"

    def whereCondition(self, params: list):
        result = 'WHERE'
        usedParamNames = []

        for condition in params:
            paramName = f"{condition['paramName']}"
            paramAlias = f"param_{condition['paramName']}"
            logicalOperator = condition['logicalOperator']
            comparisonOperator = condition['comparisonOperator']

            if isinstance(condition['value'], list):
                comparisonOperator = "IN" if comparisonOperator != "NOT IN" else comparisonOperator
                next = 0
                while True:
                    if not f"{paramAlias}_{next}" in usedParamNames:
                        break
                    next += 1
                joinedValues = []
                for i in range(len(condition['value'])):
                    inParamName = f"{paramAlias}_{i + next}"
                    usedParamNames.append(inParamName)
                    joinedValues.append(f"%({inParamName})s")
                value = f"({','.join(joinedValues)})"
                if len(condition['value']) < 1:
                    value = '1'
                    paramName = ''
                    paramAlias = ''
                    logicalOperator = ''
                    comparisonOperator = ''
            else:
                value = f" %({paramAlias})s" if condition['value'] is not None else ''

            if logicalOperator is not None:
                result = f"{result} {logicalOperator}"

            result = f"{result} {paramName} {comparisonOperator}{value}"

        return result



import mysql.connector

class Cnn:
    def __init__(self, host: str, database: str, user: str, password: str, port: int=3306, autocommit = True):
        self.autocommit = autocommit
        
        self.cnn = mysql.connector.connect(
            host=host,
            port=port,
            user=user,
            password=password,
            database=database
        )

        self.dialect = 'mysql'

    def __del__(self):
        try:
            self._cursor.close()
            self.cnn.close()
        except Exception as exc:
            pass

    def create(self, sql: str, data):
        self._cursor = self.cnn.cursor(buffered=True, dictionary=True)

        if isinstance(data, list):
            self._cursor.executemany(sql, data)
        
        elif isinstance(data, dict):
            self._cursor.execute(sql, data)

        id = self._cursor.lastrowid
        
        if self.autocommit:
            self.commit()

        return id

    def read(self, sql: str, params: dict = {}, onlyFirstRow: bool = False):
        self._cursor = self.cnn.cursor(buffered=True, dictionary=True)
        self._cursor.execute(sql, params)
        if onlyFirstRow: 
            result = self._cursor.fetchone()
        else:
            result = self._cursor.fetchall()
        self._cursor.close()
        return result
    
    def update(self, sql: str, mysqlParams: dict):
        self._cursor = self.cnn.cursor(buffered=True, dictionary=True)
        self._cursor.execute(sql, mysqlParams)
        affectedRows = self._cursor.rowcount
        
        if self.autocommit:
            self.commit()

        return affectedRows
    
    def delete(self, sql: str, params: dict = {}):
        self._cursor = self.cnn.cursor(buffered=True, dictionary=True)
        self._cursor.execute(sql, params)
        affectedRows = self._cursor.rowcount
        
        if self.autocommit:
            self.commit()

        return affectedRows
    
    def commit(self):
        self.cnn.commit()
        self._cursor.close()

    def rollback(self):
        self.cnn.rollback()
        self._cursor.close()



class SqlBuilder:
    def __init__(self, table: str):
        self.table = table
        self.basicSelect = f'SELECT * FROM `{table}`'
        self.primaryKeyQuery = f"SHOW KEYS FROM {table} WHERE Key_name = 'PRIMARY'"

    def insert(self, data):
        if isinstance(data, dict):
            keys = list(data.keys())
        elif isinstance(data, list):
            keys = list(data[0].keys())

        values = []
        for key in keys:
            values.append(f"%({key})s")

        sql = f"INSERT INTO `{self.table}` ({','.join(keys)}) VALUES({','.join(values)})"
        return sql
    
    def update(self, data: dict):
        pairs = []
        for key in data:
            pairs.append(f"{key} = %({key})s")

        return f"UPDATE `{self.table}` SET {','.join(pairs)}"
    
    def delete(self):
        return f"DELETE FROM `{self.table}`"
    
    def whereCondition(self, params: list):
        result = 'WHERE'
        usedParamNames = []

        for condition in params:
            # Define param name:
            paramName = f"{condition['paramName']}"
            paramAlias = f"param_{condition['paramName']}"

            # Define logical operator:
            logicalOperator = condition['logicalOperator']
            
            # Define comparison operator:
            comparisonOperator = condition['comparisonOperator']

            # Define condition's value:
            if isinstance(condition['value'], list):
                comparisonOperator = "IN" if comparisonOperator != "NOT IN" else comparisonOperator

                # Find next:
                next = 0
                while True:
                    if not f"{paramAlias}_{next}" in usedParamNames:
                        break
                    
                    next = next + 1

                joinedValues = []
                for i in range(0, len(condition['value'])):
                    inParamName = f"{paramAlias}_{i + next}"
                    usedParamNames.append(inParamName)
                    joinedValues.append(f"%({inParamName})s")

                value = f"({','.join(joinedValues)})"

                if len(condition['value']) < 1: 
                    value = '1'
                    paramName = ''
                    paramAlias = ''
                    logicalOperator = ''
                    comparisonOperator = ''

            else: 
                value = f" %({paramAlias})s" if condition['value'] is not None else ''

            # Set logical operator:
            if not logicalOperator == None:
                result = f"{result} {logicalOperator}"

            result = f"{result} {paramName} {comparisonOperator}{value}"

        return result



import hashlib
from importlib import import_module
from typing import Callable

class GetDao:
    def __init__(self, tablename:str, cnn):
        self.table = tablename
        self.filters = []
        self.persistence = {}
        
        if not hasattr(cnn, 'dialect'):
            raise Exception("The Cnn instance must expose a 'dialect' attribute")
        
        self.cnn = cnn
        self.dialect = cnn.dialect

        builder_module = import_module(f"sdao.{self.dialect}.sqlbuilder")
        self.sqlbuilder = builder_module.SqlBuilder(tablename)
    
    def find(self, sql: str = '', debug = False, onlyFirstRow = False):
        userDefinedParams = True
        if sql == '':
            sql = f"{self.sqlbuilder.basicSelect}{(' '+self.sqlbuilder.whereCondition(self.filters)) if len(self.filters) > 0 else ''}"
            userDefinedParams = False

        if debug: return {'SQL':sql, 'Params': self.prepareParams(userDefinedParams)}

        queryHash = self.findQueryHash(sql)
        if not queryHash in self.persistence:
            self.persistence[queryHash] = self.cnn.read(sql, self.prepareParams(userDefinedParams), onlyFirstRow)

        result = self.persistence[queryHash]
        return result
    
    def first(self, sql: str = '', debug = False):
        return self.find(sql, debug, True)
    
    def fetch(self, callback: Callable, sql: str = '', debug = False):
        result = self.find(sql, debug)

        if not debug:
            for row in result:
                callback(row)
        
        return result

    def insert(self, data, debug = False):
        sql = self.sqlbuilder.insert(data)

        if debug: return {'SQL':sql, 'Data': data}

        lastId = self.cnn.create(sql, data)

        tbKeyName = self.findTableKey()
        if isinstance(data, dict):
            data[tbKeyName] = lastId
        elif isinstance(data, list):
            for obj in data:
                obj[tbKeyName] = lastId
                lastId = lastId + 1

        return data
    
    def update(self, data: dict, debug = False):
        sql = f"{self.sqlbuilder.update(data)}{(' '+self.sqlbuilder.whereCondition(self.filters)) if len(self.filters) > 0 else ''}"

        sqlParams = data
        sqlParams.update(self.prepareParams())
        if debug: 
            return {'SQL':sql, 'Params': sqlParams}

        affectedRows = self.cnn.update(sql, sqlParams)
        return affectedRows

    def delete(self, debug = False):
        sql = f"{self.sqlbuilder.delete()}{(' '+self.sqlbuilder.whereCondition(self.filters)) if len(self.filters) > 0 else ''}"

        if debug: return {'SQL':sql, 'Params': self.prepareParams()}

        affectedRows = self.cnn.delete(sql, self.prepareParams())
        return affectedRows

    def filter(self, paramName: str):
        self.filters.append({
            'paramName': paramName,
            'logicalOperator': None,
            'comparisonOperator': None,
            'value': None
        })

        return self
    
    def _and(self, paramName: str):
        if len(self.filters) == 0:
            raise Exception(
                "You can only call this method after calling 'filter()' first.")
        
        self.filters.append({
            'paramName': paramName,
            'logicalOperator': 'AND',
            'comparisonOperator': None,
            'value': None
        })

        return self
    
    def _or(self, paramName: str):
        if len(self.filters) == 0:
            raise Exception(
                "You can only call this method after calling 'filter()' first.")
        
        self.filters.append({
            'paramName': paramName,
            'logicalOperator': 'OR',
            'comparisonOperator': None,
            'value': None
        })

        return self

    def equalsTo(self, value):
        i = len(self.filters)
        if i == 0 or self.filters[i - 1]['value'] != None:
            raise Exception("This method can only be called right after one of the filtering methods.")
        
        i = i - 1

        self.filters[i]['value'] = value
        self.filters[i]['comparisonOperator'] = 'IS NULL' if value == None else '='

        return self

    def notEqualsTo(self, value):
        i = len(self.filters)
        if i == 0 or self.filters[i - 1]['value'] != None:
            raise Exception("This method can only be called right after one of the filtering methods.")
        
        i = i - 1

        self.filters[i]['value'] = value
        self.filters[i]['comparisonOperator'] = 'IS NOT NULL' if value == None else '!='

        return self

    def biggerThan(self, value):
        i = len(self.filters)
        if i == 0 or self.filters[i - 1]['value'] != None:
            raise Exception("This method can only be called right after one of the filtering methods.")
        
        i = i - 1

        self.filters[i]['value'] = value
        self.filters[i]['comparisonOperator'] = '>'

        return self

    def lessThan(self, value):
        i = len(self.filters)
        if i == 0 or self.filters[i - 1]['value'] != None:
            raise Exception("This method can only be called right after one of the filtering methods.")
        
        i = i - 1

        self.filters[i]['value'] = value
        self.filters[i]['comparisonOperator'] = '<'

        return self

    def biggerOrEqualsTo(self, value):
        i = len(self.filters)
        if i == 0 or self.filters[i - 1]['value'] != None:
            raise Exception("This method can only be called right after one of the filtering methods.")
        
        i = i - 1

        self.filters[i]['value'] = value
        self.filters[i]['comparisonOperator'] = '>='

        return self

    def lessOrEqualsTo(self, value):
        i = len(self.filters)
        if i == 0 or self.filters[i - 1]['value'] != None:
            raise Exception("This method can only be called right after one of the filtering methods.")
        
        i = i - 1

        self.filters[i]['value'] = value
        self.filters[i]['comparisonOperator'] = '<='

        return self

    def like(self, value):
        i = len(self.filters)
        if i == 0 or self.filters[i - 1]['value'] != None:
            raise Exception("This method can only be called right after one of the filtering methods.")
        
        i = i - 1

        self.filters[i]['value'] = f"%{value}%"
        self.filters[i]['comparisonOperator'] = 'LIKE'

        return self

    def _in(self, value: list):
        i = len(self.filters)
        if i == 0 or self.filters[i - 1]['value'] != None:
            raise Exception("This method can only be called right after one of the filtering methods.")
        
        i = i - 1

        self.filters[i]['value'] = value
        self.filters[i]['comparisonOperator'] = 'IN'

        return self

    def _notIn(self, value: list):
        i = len(self.filters)
        if i == 0 or self.filters[i - 1]['value'] != None:
            raise Exception("This method can only be called right after one of the filtering methods.")
        
        i = i - 1

        self.filters[i]['value'] = value
        self.filters[i]['comparisonOperator'] = 'NOT IN'

        return self
    
    def prepareParams(self, userDefinedParams: bool = False):
        params = {}
        for condition in self.filters:
            if isinstance(condition['value'], list):
                # Find next:
                next = 0
                while True:
                    if not f"{'param_' if not userDefinedParams else '' }{condition['paramName']}_{next}" in params:
                        break
                    
                    next = next + 1

                for i in range(0, len(condition['value'])):
                    val = condition['value'][i]
                    params[f"{'param_' if not userDefinedParams else '' }{condition['paramName']}_{i + next}"] = val

            else: params[f"{'param_' if not userDefinedParams else '' }{condition['paramName']}"] = condition['value']

        return params
    
    def findQueryHash(self, sql: str):
        result = sql.replace('param_', '')
        params = self.prepareParams()

        for key in params:
            val = params[key]
            key = key.replace('param_', '')
            if isinstance(val, list):
                for i in range(0, len(val)):
                    result = result.replace(f"%({key}_{i})s", str(val[i]))
            else:
                result = result.replace(f"%({key})s", str(val))

        result = hashlib.md5(result.encode())
        return result.hexdigest()

    def findTableKey(self):
        sql = self.sqlbuilder.primaryKeyQuery
        tbinfo = self.first(sql)
        return tbinfo['Column_name'] if tbinfo else None



from setuptools import setup, find_packages

setup(
    name="sdao",
    version="1.1.1",
    description="A DAO (Data Access Object) library for Python",
    author="Gabriel Valentoni Guelfi",
    author_email="gabriel.valguelfi@gmail.com",
    packages=find_packages(),  # Encontra sdao/
    python_requires=">=3.7",
    install_requires=[],
    extras_require={
        "mysql": ["mysql-connector-python"],
        "mariadb": ["mariadb"],
        "mssql": ["pyodbc"]
    },
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import mariadb

class Cnn:
    def __init__(self, host: str, database: str, user: str, password: str, port: int = 3306, useSocket = False, autocommit = True):
        self.autocommit = autocommit

        if host == 'localhost' and not useSocket:
            host = '127.0.0.1'

        self.cnn = mariadb.connect(
            host=host,
            port=port,
            user=user,
            password=password,
            database=database
        )

        self.dialect = 'mariadb'
        
    def __del__(self):
        try:
            self._cursor.close()
            self.cnn.close()
        except Exception as exc:
            pass

    def _fetchall_as_dicts(self, cursor):
        columns = [col[0] for col in cursor.description]
        return [dict(zip(columns, row)) for row in cursor.fetchall()]

    def _fetchone_as_dict(self, cursor):
        columns = [col[0] for col in cursor.description]
        row = cursor.fetchone()
        return dict(zip(columns, row)) if row else None

    def create(self, sql: str, data):
        self._cursor = self.cnn.cursor()
        if isinstance(data, list):
            self._cursor.executemany(sql, data)
        elif isinstance(data, dict):
            self._cursor.execute(sql, data)

        last_id = self._cursor.lastrowid

        if self.autocommit:
            self.commit()

        return last_id

    def read(self, sql: str, params: dict = {}, onlyFirstRow: bool = False):
        self._cursor = self.cnn.cursor()
        self._cursor.execute(sql, params)
        result = self._fetchone_as_dict(self._cursor) if onlyFirstRow else self._fetchall_as_dicts(self._cursor)
        self._cursor.close()
        return result

    def update(self, sql: str, params: dict):
        self._cursor = self.cnn.cursor()
        self._cursor.execute(sql, params)
        affectedRows = self._cursor.rowcount

        if self.autocommit:
            self.commit()

        return affectedRows

    def delete(self, sql: str, params: dict = {}):
        self._cursor = self.cnn.cursor()
        self._cursor.execute(sql, params)
        affectedRows = self._cursor.rowcount

        if self.autocommit:
            self.commit()
            
        return affectedRows
    
    def commit(self):
        self.cnn.commit()
        self._cursor.close()

    def rollback(self):
        self.cnn.rollback()
        self._cursor.close()



import pyodbc
import re

class Cnn:
    def __init__(self, host: str, database: str, user: str, password: str, port: str="1433", autocommit = True):
        self.autocommit = autocommit
        
        self.cnn = pyodbc.connect(
            'DRIVER={ODBC Driver 17 for SQL Server};'
            f'SERVER={host},{port};'
            f'DATABASE={database};'
            f'UID={user};'
            f'PWD={password}'
        )

        self.dialect = 'mssql'

    def __del__(self):
        try:
            self._cursor.close()
            self.cnn.close()
        except Exception as exc:
            pass

    def _fetchall_as_dicts(self, cursor):
        columns = [col[0] for col in cursor.description]
        return [dict(zip(columns, row)) for row in cursor.fetchall()]

    def _fetchone_as_dict(self, cursor):
        columns = [col[0] for col in cursor.description]
        row = cursor.fetchone()
        return dict(zip(columns, row)) if row else None

    def create(self, sql: str, data):
        self._cursor = self.cnn.cursor()

        if isinstance(data, list):
            self._cursor.executemany(sql, data)
        elif isinstance(data, dict):
            self._cursor.execute(sql, data)

        # pyodbc does not support lastrowid reliably across all backends
        # Use SCOPE_IDENTITY() explicitly for SQL Server if needed
        self._cursor.execute("SELECT SCOPE_IDENTITY()")
        last_id = self._cursor.fetchone()[0]

        if self.autocommit:
            self.commit()

        return last_id

    def read(self, sql: str, params: dict = {}, onlyFirstRow: bool = False):
        self._cursor = self.cnn.cursor()

        # detecta se a query tem marcadores no estilo %(param)s
        has_markers = bool(re.search(r"%\([^)]+\)s", sql))
        has_params = bool(params)

        if has_markers and has_params:
            self._cursor.execute(sql, params)
        else:
            self._cursor.execute(sql)

        result = self._fetchone_as_dict(self._cursor) if onlyFirstRow else self._fetchall_as_dicts(self._cursor)
        self._cursor.close()
        return result

    def update(self, sql: str, params: dict):
        self._cursor = self.cnn.cursor()

        # detecta se a query tem marcadores no estilo %(param)s
        has_markers = bool(re.search(r"%\([^)]+\)s", sql))
        has_params = bool(params)

        if has_markers and has_params:
            self._cursor.execute(sql, params)
        else:
            self._cursor.execute(sql)
            
        affectedRows = self._cursor.rowcount
        
        if self.autocommit:
            self.commit()

        return affectedRows

    def delete(self, sql: str, params: dict = {}):
        self._cursor = self.cnn.cursor()

        # detecta se a query tem marcadores no estilo %(param)s
        has_markers = bool(re.search(r"%\([^)]+\)s", sql))
        has_params = bool(params)

        if has_markers and has_params:
            self._cursor.execute(sql, params)
        else:
            self._cursor.execute(sql)

        affectedRows = self._cursor.rowcount
        
        if self.autocommit:
            self.commit()

        return affectedRows
    
    def commit(self):
        self.cnn.commit()
        self._cursor.close()

    def rollback(self):
        self.cnn.rollback()
        self._cursor.close()
